############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time, types, warnings

import ZSI
from ZSI import EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.fault import FaultFromFaultMessage, FaultType
from ZSI.address import Address, WSActionException
from ZSI.writer import SoapWriter
from ZSI.wstools.Namespaces import OASIS

from twisted.web import client
from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log, failure
from twisted.python.failure import Failure
from twisted.internet.tcp import *

from pyGridWare.faults import WrapFaultInException
try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex:
    DomletteReader = DomletteElementProxy = None


class HTTPPageGetter(client.HTTPPageGetter):
    def handleStatus_500(self):
        '''potentially a SOAP:Fault.
        '''
        pass
    def handleStatus_404(self):
        '''client error, not found
        '''
        log.debug('HTTP Error 404')

try:
    from OpenSSL import SSL
    SSL.Session
except:
    client.HTTPClientFactory.protocol = HTTPPageGetter
else:
    class SMHTTPPageGetter(HTTPPageGetter):
        """Retains the SSL session
        """
        session = None

        def makeConnection(self, transport):
            """Do handshake and set session id.
            To make this a little more usefull need to
            associate each session with an URL and security context..
            """
            socket = transport.getHandle()
            use_session = hasattr(socket, 'set_session')
            session = SMHTTPPageGetter.session
            if use_session is True:
                if session is None:
                    SMHTTPPageGetter.session = SSL.Session(socket)
                else:
                    socket.set_session(session)

            client.HTTPPageGetter.makeConnection(self, transport)

    client.HTTPClientFactory.protocol = SMHTTPPageGetter
    


def getPage(url, contextFactory=None, *args, **kwargs):
    """Download a web page as a string.

    Download a page. Return a deferred, which will callback with a
    page (as a string) or errback with a description of the error.

    See HTTPClientFactory to see what extra args can be passed.
    """
    scheme, host, port, path = client._parse(url)
    factory = client.HTTPClientFactory(url, *args, **kwargs)
    if scheme == 'https':
        if contextFactory is None:
            raise RuntimeError, 'must provide a contextFactory'
        conn = reactor.connectSSL(host, port, factory, contextFactory)
    else:
        conn = reactor.connectTCP(host, port, factory)

    return factory


class Binding:
    def __init__(self, url=None, nsdict=None, readerclass=DomletteReader, 
                 writerclass=DomletteElementProxy, wsAddressURI=None, 
                 sig_handler=None, contextFactory=None, **kw):

        self.url = url
        self.nsdict = nsdict or {}
        self.readerclass = readerclass
        self.writerclass = writerclass
        self.wsAddressURI = wsAddressURI
        self.sig_handler = sig_handler
        self.endPointReference = kw.get('endPointReference')
        self.contextFactory = contextFactory

        self._factory = None
        self._agent = 'pyGridWare Binding'
        self.http_headers  = {'content-type': 'text/xml',}
        self._fault = False
        self.failure = None

    def addHTTPHeader(self, key, value):
        self.http_headers[key] = value
   
    def _getHTTPHeaders(self):
        return self.http_headers
        
    def Send(self, url, opname, obj, nsdict={}, soapaction=None, **kw):
        '''serialize obj and send 
        '''
        url = url or self.url
        cookies = None
        if self._factory is not None:
            cookies = self._factory.cookies
         
        self.addHTTPHeader('SOAPAction', soapaction)
        address,soapdata = self._getRequestSOAPMessage(url, obj, nsdict, soapaction, **kw)
        f = getPage(str(url), contextFactory=self.contextFactory, 
                          postdata=soapdata, agent=self._agent, 
                          method='POST', headers=self._getHTTPHeaders(), cookies=cookies)
        if isinstance(f, Failure):
            return f
        d = f.deferred
        d.address = address
        # Hold on to factory for deferred and cookies.
        self._factory = f
        return d

    def Receive(self, replytype, **kw):
        '''This method allows code to act in a synchronous manner, it waits to return until 
        the deferred fires but it doesn't prevent other queued called from being executed.
        Send must be called first, which sets up the factory.
        '''
        assert getattr(self, '_factory') is not None, 'Must call Send before Receive.'
        assert isinstance(replytype, ZSI.TC.TypeCode), 'replytype must be a TypeCode.'

        self.failure = None
        d = self._factory.deferred
        d.addCallback(self.getResponseSOAPMessage, replytype=replytype, address=d.address, **kw)
        d.addErrback(self.errback)

        while not d.called:
            reactor.runUntilCurrent()
            t2 = reactor.timeout()
            t = reactor.running and t2
            reactor.doIteration(t)

        pyobj = d.result
        if isinstance(pyobj, Exception) is True:
            raise pyobj

        return pyobj

    def errback(self, result):
        '''used with Response method.
        '''
        log.debug('Errback called: %s', result)
        self.failure = result
        return result

    def getResponseSOAPMessage(self, soapdata, replytype, **kw):
        '''called by deferred, returns pyobj representing reply.
        Parameters and Key Words:
          soapdata -- SOAP Data
          replytype -- reply type of response
          address -- WS-Address
        '''
        address = kw.get('address')
        if address is None:
            raise EvaluateException, 'Missing WS-Address'
        
        if len(soapdata) == 0:
            raise TypeError('Received empty response')

        log.debug("_" * 33, time.ctime(time.time()), "RESPONSE: \n%s" %soapdata)

        ps = ParsedSoap(soapdata, readerclass=self.readerclass)
        if ps.IsAFault() is True:
            raise FaultFromFaultMessage(ps)

        pyobj = ps.Parse(replytype)
        wsaction = kw.get('wsaction')
        if wsaction is not None:
            address.checkResponse(ps, wsaction)

        if self.sig_handler is not None:
            self.sig_handler.verify(ps)

        return pyobj

    def _getRequestSOAPMessage(self, url, obj, nsdict={}, soapaction=None, **kw):
        '''returns string representing request
        '''
        url = url or self.url
        if kw.has_key('requesttypecode'):
            tc = kw['requesttypecode']
        elif kw.has_key('requestclass'):
            tc = kw['requestclass'].typecode
        elif type(obj) == types.InstanceType:
            tc = getattr(obj.__class__, 'typecode')
        elif hasattr(obj, 'typecode'):
            tc = getattr(obj.__class__, 'typecode')
        else:
            raise RuntimeError, 'Must provide a typecode to serialize obj'

        # Serialize the object.
        d = {}
        d.update(self.nsdict)
        d.update(nsdict)

        use_header = self.sig_handler is not None or self.wsAddressURI is not None
        sw = SoapWriter(nsdict=d, header=use_header, outputclass=self.writerclass)
        sw.serialize(obj, tc)

        # Serialize WS-Address
        address = None
        if self.wsAddressURI:
            epr = kw.get('endPointReference', self.endPointReference)
            wsaction = kw.get('wsaction')
            if soapaction and wsaction.strip('\'"') != soapaction:
                raise WSActionException, 'soapAction(%s) and WS-Action(%s) must match'\
                    %(soapaction,wsaction)

            address = Address(url, wsAddressURI=self.wsAddressURI)
            address.setRequest(epr, wsaction)
            address.serialize(sw)

        if self.sig_handler is not None:
            self.sig_handler.sign(sw)
        soapdata = str(sw)

        log.debug("_" * 33, time.ctime(time.time()), "REQUEST: \n%s" %soapdata)

        return address,soapdata


