###########################################################################
# Monte M. Goode, LBNL
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

# $Id: resource.py,v 1.24 2005/06/24 08:52:16 boverhof Exp $

import os,sys,warnings
from cStringIO import StringIO
from twisted import copyright
TWISTED_VERSION = (2,0,0)
try:
    if tuple(map(lambda i:int(i), copyright.version.split('.'))) < TWISTED_VERSION:
        warnings.warn(\
            'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
            RuntimeWarning,
        )
except:
    warnings.warn(\
        'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
        RuntimeWarning,
    )

from zope.interface import Interface
from twisted.web.error import NoResource
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource
from twisted.python import log, failure

from ZSI import _get_element_nsuri_name, EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.writer import SoapWriter
from ZSI import fault
from ZSI.address import Address
from ZSI.ServiceContainer import WSActionException, WSActionNotSpecified

try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex: 
    DomletteReader = DomletteElementProxy = None
    
from pyGridWare.faults import BaseFault
from pyGridWare.utility.http import GetURLFromRequest


class FileServerResource(Resource, object):
    '''Represents a root directory to serve files out of.
    '''
    def __init__(self, root='wsdl'):
        '''root -- root directory path.
        '''
        self.root = root
        Resource.__init__(self)

    def render_GET(self, request):
        s = StringIO()
        if request.path.startswith('/%s' %self.root) is False:
            raise NoResource, request.path

        rpath = request.path.split(self.root, 1)[1].lstrip('/')
        fpath = os.path.join(self.root, rpath)
        if os.path.isfile(fpath) is True:
            for l in open(fpath,'r'): s.write(l)
            return s.getvalue()

        raise NoResource, 'request(%s), no file(%s)' %(request.path,fpath)

    def getChild(self, path, request):
        '''so can serve GET requests and act as directory.
        '''
        return self


class IWSSecurity(Interface):
    '''Interface for WS-Security'''

    def sign(self, sw):
        '''Add XML Signatures'''

    def verify(self, ps):
        '''Verify Signatures'''


class IWSSubscriptionManager(Interface):
    '''Interface for WS-BaseNotification SubscriptionManager'''

    def subscribe(self, msg, ctx, request):
        '''Subscribe to EPR specified in msg
        msg -- SubscriptionRequest instance
        ctx -- ResourceContext instance
        request -- twisted.http.Request instance
        '''



class WSResource(Resource, object):
    '''
    class variables:
        encoding  
        soapAction -- Must override to set input Action values.
           Dictionary of soapAction keys, and operation name values.
           These are specified in the WSDL soap bindings. There must be a 
           class method matching the operation name value.  If WS-Action is
           used the keys are WS-Action request values, according to the spec
           if soapAction and WS-Action is specified they must be equal.
           
        wsAction -- Must override to set output Action values.
           Dictionary of operation name keys and WS-Action 
           response values.  These values are specified by the portType.

        root -- dictionary of root element keys, and operation name values.
    '''
    soapAction = {}
    wsAction = {}
    root = {}
    encoding = "UTF-8"

    def __init__(self):
        Resource.__init__(self)

    def render_POST(self, request):
        """Dispatch Method called by twisted render.
        request -- twisted.web.server.Request
        """
        http_headers = request.getAllHeaders()
        soap_action = http_headers.get('soapaction')
        log.debug('Request soapAction -- %s' %soap_action)
        data = request.content.read()
        log.debug(data)
        try:
            ps = ParsedSoap(data, readerclass=DomletteReader)
        except ParseException, e:
            return self._soapFault(request, e)

        # Verify if Signed
        if IWSSecurity.providedBy(self):
            self.verify(ps)

        # WS-Address
        address = Address()
        try:
            address.parse(ps)
        except ParseException, e:
            return self._soapFault(request, e)

        action = address.getAction()
        if not action:
            log.err('No WS-Action specified in Request')
            e = WSActionException('No WS-Action specified in Request')
            return self._soapFault(request, e)

        if soap_action and soap_action.strip('\'"') != action:
            log.err(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.'\
                %(soap_action,action)
            )
            e = WSActionException(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.'\
                %(soap_action,action)
            )
            return self._soapFault(request, e)

        op_name = self.root.get(_get_element_nsuri_name(ps.body_root)) or \
            self.soapAction.get(action)
        if op_name is None:
            raise UnknownRequestException,\
                'failed to map request to a method: action(%s), root%s'\
                %(action,_get_element_nsuri_name(ps.body_root))
        
        try:
            method = getattr(self, op_name)
            result = method(ps, address, request=request)
        except Exception, e:
            log.err('Failed to locate service instance/bound method from WS-Address')
            return self._soapFault(request, e)

        if result is None:
            log.debug('No Response specified')
            return NOT_DONE_YET

        sw = SoapWriter(outputclass=DomletteElementProxy)
        try:
            sw.serialize(result)
        except Exception, e:
            return self._soapFault(request, e)

        if self.wsAction.has_key(op_name) is False:
            raise WSActionNotSpecified,\
                'wsAction dictionary missing key(%s)' %op_name
            
        rsp_action = self.wsAction[op_name]
        addressRsp = Address(action=rsp_action)
        addressRsp.setResponseFromWSAddress(address, GetURLFromRequest(request))
        try:
            addressRsp.serialize(sw)
        except Exception, e:
            log.err('Problem serializing WS-Address for Response')
            return self._soapFault(request, e)
            
        # Sign if WS-Security
        if IWSSecurity.providedBy(self):
            self.sign(sw)
            
        soapdata = str(sw)
        self._sendResponse(request, soapdata)
        return NOT_DONE_YET

    def _sendResponse(self, request, response, status=200):
        '''
        request -- request message
        response --- response message
        status -- HTTP Status
        '''
        request.setResponseCode(status)

        if self.encoding is not None:
            mimeType = 'text/xml; charset="%s"' % self.encoding
        else:
            mimeType = "text/xml"

        request.setHeader("Content-type", mimeType)
        request.setHeader("Content-length", str(len(response)))
        request.write(response)
        request.finish()

    def _soapFault(self, request, e):
        '''
        request -- request message
        e -- Exception 
        '''
        response = None
        if isinstance(e, BaseFault):
            response = fault.Fault(fault.Fault.Server, 'userException', None, 
                [e.fault], None).AsSoap()
        else:
            response = fault.FaultFromException(e, False, sys.exc_info()[2]).AsSOAP()

        log.err('SOAP FAULT: %s' % response)
        self._sendResponse(request, response, status=500)
        return NOT_DONE_YET

