##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, warnings
from distutils import log
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from ConfigParser import ConfigParser, NoOptionError
from ZSI.wstools.WSDLTools import WSDLReader

sys.path.append('%s/pyGridWare/utility/generate' %os.getcwd())
from Modules import WriteServiceModules, WriteClientStubModule,\
    WriteTypesModule, NR
from Utility import CLIENT,SERVICE,SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE

CONFIG_FILE = 'config.txt'
CACHE_FILE = '.pyGridWare.cache'
CACHE_SECTION = 'distribution'
CACHE_OPTION = 'packages'
WSDL_SEC = 'WSDL'

PERSISTENT_SEC = 'Persistent'

def RegenerateModules(items, services=True, stubs=True, cp=None):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return

    add_packages = NR.Reset()
    add_packages.insert(0, 'pyGridWare.generated')
    for opt,url in items:
        log.info("Generate: %s, %s",opt,url)
        wsdl = None
        reader = WSDLReader()
        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        kw = {}
        if cp is not None:
            try: kw['simple'] = not cp.getboolean(PERSISTENT_SEC, opt)
            except NoOptionError: pass
            
        if services:
            WriteServiceModules(opt, wsdl, **kw)
        if stubs:
            WriteClientStubModule(opt, wsdl)
        else:
            WriteTypesModule(opt, wsdl)

    return add_packages


class build(_build):
    '''Specialized build.
    '''
    user_options = _build.user_options[:]
    user_options.append(\
        ('generated', 'g',
         'create generated modules/directories and build cache file "%s"' %CACHE_FILE)
    )
    user_options.append(\
        ('force', 'f',
         'write generated modules into restricted packages')
    )
    user_options.append(\
        ('options=', 'o',
         'restrict generation to whitespace separated list of option(s) from config.txt section %s' %WSDL_SEC)
    )

    boolean_options = _build.boolean_options[:]
    boolean_options.append('generated')

    def initialize_options(self):
        _build.initialize_options(self)
        self.generated = None
        self.force = None
        self.options = None

    def run(self):
        '''if generate is specified generate modules.
        '''
        if not self.generated:
            return _build.run(self)

        if self.force is None:
            bad = filter(lambda p: p in self.distribution.packages, map(NR.GetModFromFile, NR.d.values()))
            if len(bad) > 0:
                log.warn("force to generate modules into restricted package %s" %bad)
                sys.exit(-1)

        cp = ConfigParser()
        cp.read(CONFIG_FILE)
        items = None
        if self.options is not None:
            if isfile(CACHE_FILE) is False:
                raise RuntimeError, 'Must build all generated modules first, remove --options.'
 
            items = []
            options = cp.options(WSDL_SEC)
            for opt in map(lambda o: o.lower(), self.options.split()):
                if opt in options: 
                    items.append((opt,cp.get(WSDL_SEC,opt)))
                else:
                    warnings.warn('IGNORING option %s, it is not in section %s' %(opt,WSDL_SEC))

        else:
            items = cp.items(WSDL_SEC)

        packages = RegenerateModules(items, services=True, stubs=True, cp=cp)
        del cp
        try:
            packages.remove('pyGridWare')
        except ValueError, ex:
            pass

        self.distribution.packages += packages
        ok = _build.run(self)
        if self.options is not None:
            return ok

        cp = ConfigParser()
        fd = open(CACHE_FILE, 'w+')
        cp.add_section(CACHE_SECTION)
        cp.set(CACHE_SECTION, CACHE_OPTION, packages)
        cp.write(fd)
        fd.close()
        del cp
        return ok


class clean(_clean):
    '''Specialized clean.
    '''
    user_options = _clean.user_options[:]
    user_options.append(\
        ('generated', 'g',
         'remove all generated modules/directories')
    )
    user_options.append(\
        ('force', 'f',
         'force removal of all generated modules/directories')
    )

    boolean_options = _clean.boolean_options[:]
    boolean_options.append('generated')

    def initialize_options(self):
        _clean.initialize_options(self)
        self.generated = None
        self.force = None

    def run(self):
        '''if generated is specified remove all such files.
        '''
        _clean.run(self)
        if not self.generated: return

        func1 = lambda f: f != '__init__.py' and f.endswith('.py')
        files = []

        if os.path.isfile(CACHE_FILE):
            cp = ConfigParser()
            cp.read(CACHE_FILE)
            packages = eval(cp.get(CACHE_SECTION, CACHE_OPTION))
            for p in packages:
                if p.strip() == 'pyGridWare': 
                    log.warn("Will not remove directory pyGridWare")
                    continue
                if self.force is None and p.startswith('pyGridWare.generated') is False:
                    log.warn("'%s' is not in pyGridWare/generated -- try force", dir)
                    break

                dir = p.replace('.', '/')
                if os.path.exists(dir):
                    log.warn("REMOVE %s" %dir)
                    remove_tree(dir, dry_run=self.dry_run)
            else:
                log.warn("REMOVE %s" %CACHE_FILE)
                os.remove(CACHE_FILE)
            del cp


class PyGridWareDist(Distribution): 
    global_options = Distribution.global_options[:]
    global_options.append(('regenerate', None, "Deprecated, use build command with generated flag"))

    def parse_command_line(self):
        ok = Distribution.parse_command_line(self)
        if self.verbose == 1:
            from ZSI.wstools.logging import setBasicLoggerDEBUG
            setBasicLoggerDEBUG()

        if '--regenerate' in self.script_args:
            raise DeprecationWarning, \
                'regenerate global option no longer supported.  Use build command.'

        cp = ConfigParser()
        if os.path.isfile(CACHE_FILE) is True:
            cp = ConfigParser()
            cp.read(CACHE_FILE)
            self.packages += eval(cp.get(CACHE_SECTION, CACHE_OPTION))
            del cp
        else:
            warnings.warn(\
                '***No Modules have been generated ("%s")\n\tpython setup.py build -g' %CACHE_FILE,
            )
        return ok 


def main():
    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Programming Language :: Python
        Natural Language :: English
        Operating System :: Microsoft :: Windows
        Operating System :: POSIX
        """

    longDesc = """
        Python implementation of the WS-Resource Framework (WSRF).  
        pyGridWare has facilities for secure messaging including security 
        context negotiation, delegation, and GSS XML Signatures.
        """

    setup(name="pyGridWare", 
          version="1.0.2beta", 
          packages=["pyGridWare", 
                    "pyGridWare.addressing", 
                    "pyGridWare.lifetime", 
                    "pyGridWare.notification",
                    "pyGridWare.plugins",
                    "pyGridWare.properties",
                    "pyGridWare.faults",
                    "pyGridWare.resource", 
                    "pyGridWare.security", 
                    "pyGridWare.security.gss", 
                    "pyGridWare.security.base",
                    "pyGridWare.services", 
                    "pyGridWare.utility", 
                    "pyGridWare.utility.generate", 
                    "pyGridWare.utility.scripts",
                    "pyGridWare.wsrf", 
          ],
          scripts=["tools/WSDLBuilder"],
          description="Python WS-Resource Framework and GSS Security",
	  author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
	  url="http://www-itg.lbl.gov/gtg/pyGridWare/",
          classifiers = filter(None, classifiers.split("\n"))[:-1],
          long_description=longDesc,
          distclass=PyGridWareDist,
          cmdclass={'clean':clean, 'build':build},
    )


if __name__ == '__main__': 
    main()
