#!/usr/bin/env python
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, signal, time, warnings, unittest
from distutils import log
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from ConfigParser import ConfigParser

from twisted.internet import reactor
from twisted.python import log

from ZSI import ZSIException
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.wstools.Namespaces import WSA200403

from pyGridWare.utility.generate.Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE, NameResolver as NR
from pyGridWare.utility.generate.Modules import WriteServiceModules,\
    WriteClientStubModule, WriteTypesModule
    
import pyGridWare.faults
    
NR.d[CLIENT] = 'gen.stubs'
NR.d[SERVICE] = 'gen.services'
NR.d[SERVICE_STUBS] = 'gen.services.stubs'
NR.d[TYPES] = 'gen.types'
NR.d[PROPERTIES] = 'gen.properties'
NR.d[RESOURCE] = 'gen.resource'

log.startLogging(sys.stdout)


def RegenerateModules(items, services=True, stubs=True):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return
    add_packages = NR.Reset()
    add_packages.insert(0, 'gen')
    for sec,url in items:
        #log.info("Generate: %s, %s",sec,url)
        log.debug("Generate: %s, %s",sec,url)
        wsdl = None
        reader = WSDLReader()
        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        if services:
            WriteServiceModules(sec, wsdl)
        if stubs:
            WriteClientStubModule(sec, wsdl)
        else:
            WriteTypesModule(sec, wsdl)

    return add_packages


class CounterTestCase(unittest.TestCase):
    def __init__(self, methodName='runTest', epr=None):
        unittest.TestCase.__init__(self, methodName)
        self.rsp = None
        self.epr = epr
        self.port = None

    def getMethodToCall(self):
        '''return method to call, must be callable and 
        takes request as argument.
        '''
        raise NotImplementedError, 'this method must be implemented'
        
    def NewRequestInstance(self):
        '''return request instance.
        '''
        raise NotImplementedError, 'this method must be implemented'
        
    def getEPR(self):
        '''if factory method, return EPR from response.
        '''
        return self.epr

    def setEPR(self, epr):
        '''set EndPointReference
        '''
        self.epr = epr

    def setUp(self):
        import gen.stubs.counter
        loc = gen.stubs.counter.CounterServiceLocator()
        self.port = loc.getCounterPortType(\
            url='http://localhost:9080/wsrf/services/CounterService',
            wsAddressURI=WSA200403.ADDRESS,
            endPointReference=self.epr,
            tracefile=sys.stdout,
        )
        self.shouldStop = True
        
    def tearDown(self):
        self.shouldStop = False
        
    def runTest(self, epr=None):
        kw = {}
        msg = self.NewRequestInstance()
        method = self.getMethodToCall()
        if msg is None:
            raise
        if method is None or not callable(method):
            raise
        if epr is not None:
            self.port.endPointReference = epr
        
        self.rsp = method(msg, **kw)


# Counter Specific
class AddTestCase(CounterTestCase):
    '''Counter Add
    '''
    def getMethodToCall(self):
        return self.port.add

    def NewRequestInstance(self):
        import gen.stubs.counter
        return gen.stubs.counter.AddInputMessage(10)

    def setUp(self):
        self.failUnless(self.epr is not None)
        CounterTestCase.setUp(self)
        
    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.AddOutputMessage))
        CounterTestCase.tearDown(self)


class CreateTestCase(CounterTestCase):
    '''WS-
    '''
    def getMethodToCall(self):
        return self.port.createCounter
        
    def getEPR(self):
        '''if factory method, return EPR from response.
        '''
        if self.rsp is None: return
        self.failUnless(getattr(self.rsp,'_EndpointReference',None) is not None, 
            'no EPR in create response.')
        return self.rsp._EndpointReference

    def NewRequestInstance(self):
        import gen.stubs.counter
        return gen.stubs.counter.CreateCounterRequest()
        
    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.CreateCounterResponse))
        CounterTestCase.tearDown(self)
        
# WS-Resource
class GetResourcePropertyTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetResourceProperty

    def NewRequestInstance(self):
        import gen.stubs.counter
        return gen.stubs.counter.GetResourcePropertyRequest(("http://counter.com","Value"))
        
    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.GetResourcePropertyResponse))
        CounterTestCase.tearDown(self)


class GetMultipleResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetMultipleResourceProperties

    def NewRequestInstance(self):
        import gen.stubs.counter
        msg = gen.stubs.counter.GetMultipleResourcePropertiesRequest()
        for qname in [("http://counter.com","Value"), 
            ("http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd","CurrentTime"),]:
            msg._ResourceProperty.append(qname)
        return msg

    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.GetMultipleResourcePropertiesResponse))
        CounterTestCase.tearDown(self)


class QueryResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.QueryResourceProperties

    def NewRequestInstance(self):
        from ZSI.wstools.Namespaces import DSIG
        import gen.stubs.counter
        msg = gen.stubs.counter.QueryResourcePropertiesRequest()
        msg._QueryExpression = filter(lambda what: what.aname=='_QueryExpression', msg.typecode.ofwhat)[0].pyclass()
        msg._QueryExpression._text  = "contains(/*/*[namespace-uri()='http://counter.com' and local-name()='operations supported'],\"Value\")"
        msg._QueryExpression._attrs = {\
            'Dialect':DSIG.XPATH,
        }
        return msg

    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp)
        CounterTestCase.tearDown(self)

    def runTest(self, epr=None):
        import gen.stubs.counter
        CounterTestCase.runTest(self, epr=None)
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp._any == False)


# -----------------------------------------------
# WS-Lifetime
# -----------------------------------------------
class DestroyTestCase(CounterTestCase):
    '''WS-
    '''
    def getMethodToCall(self):
        return self.port.Destroy

    def NewRequestInstance(self):
        import gen.stubs.counter
        return gen.stubs.counter.DestroyRequest()
        
    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.DestroyResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = True


class SetTerminationTimeTestCase(CounterTestCase):
    '''after this returns resource should be destroyed.
    '''
    def getMethodToCall(self):
        return self.port.SetTerminationTime

    def NewRequestInstance(self):
        import gen.stubs.counter
        msg = gen.stubs.counter.SetTerminationTimeRequest()
        msg._RequestedTerminationTime = time.gmtime(time.time()+2)
        return msg
        
    def tearDown(self):
        import gen.stubs.counter
        self.failUnless(isinstance(self.rsp, gen.stubs.counter.SetTerminationTimeResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = False
        
    def runTest(self, epr=None):
        '''After returns resource should be destroyed.
        '''
        CounterTestCase.runTest(self, epr)
        time.sleep(5)
        
        
# -----------------------------------------------
# WS-BaseNotification
# -----------------------------------------------
class NotificationTestCase(CounterTestCase):
    '''WS-BaseNotification
    '''

class ResourceUnknownTestCase(CounterTestCase):
    
    def __init__(self, klass, methodName='runTest', epr=None):
        unittest.TestCase.__init__(self, methodName)
        self.wrappedTest = klass(methodName=methodName, epr=epr)

    def getEPR(self):
        return self.wrappedTest.getEPR()

    def setEPR(self, epr):
        self.wrappedTest.setEPR(epr)

    def NewRequestInstance(self):
        raise NotImplementedError, 'this method is not implemented'
        
    def setUp(self):
        self.wrappedTest.setUp()
        
    def tearDown(self):
        self.shouldStop = True
        
    def runTest(self, epr=None):
        callableObj = self.wrappedTest.runTest
        try:
            callableObj(epr=epr)
        except ZSIException, ex:
            print ex.detail[0].typecode.pname
            self.failUnlessEqual(ex.detail[0].typecode.pname, 'ResourceUnknownFault')
       
DIR = dir()
class CounterTestSuite(unittest.TestSuite):
    def __init__(self, tests=()):
        wsdl = '../share/schema/core/samples/counter/counter_service.wsdl'
        packages = RegenerateModules(items=[('counter',wsdl)], services=True, stubs=True)
        unittest.TestSuite.__init__(self, tests=tests)

    def run(self, result):
        '''Factory method first.  Each factory method call will change the EPR
        for subsequent calls.
        '''
        epr = None
        for test in self._tests:
            if result.shouldStop: break
            
            test.setEPR(epr)
            test(result)
            epr = test.getEPR()
            
        return result


def WSRPCounterTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), AddTestCase(), GetResourcePropertyTestCase(), 
                GetMultipleResourcePropertiesTestCase(), DestroyTestCase()),
        )

def WSRPBrokenTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), QueryResourcePropertiesTestCase(), DestroyTestCase()),
        )
   
def WSLifetimeCounterTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), SetTerminationTimeTestCase(), ResourceUnknownTestCase(DestroyTestCase)),
        )


class MyTestProgram(unittest.TestProgram):
    def runTests(self):
        log.debug("RUN TESTS")
        if self.testRunner is None:
            self.testRunner = unittest.TextTestRunner(verbosity=self.verbosity)
        result = self.testRunner.run(self.test)
        log.debug("SUCCESS?: ", result.wasSuccessful())

        reactor.stop()
        os.kill(PID, signal.SIGTERM)


PID = None
def main():
    global PID
    PID = os.spawnlp(os.P_NOWAIT, 'twistd', 'twistd', '-noy',  '../bin/server-config.tac')
    reactor.callLater(1.00, MyTestProgram, defaultTest="WSRPCounterTestSuite")
    reactor.run()


if __name__ == '__main__': 
    main()
